from __future__ import annotations  # todo0 remove in 3.11

import functools
from dataclasses import dataclass, field
from typing import Callable, Iterable

import plotly
import plotly.basedatatypes
import plotly.graph_objects
import sympy
# noinspection PyProtectedMember
from plotly.io import _html, _kaleido

from flanautils import iterables, oss
from flanautils.models.bases import FlanaBase


def get_plotlyjs():
    """Hardcode a version of plotly in Spanish to render html."""

    with open(oss.resolve_path('flanautils/resources/plotly_es.js')) as f:
        return f.read()


_kaleido.scope.plotlyjs = oss.resolve_path('flanautils/resources/plotly_es.js')  # render image
_html.get_plotlyjs = get_plotlyjs  # render html


def find_resolution(func: Callable = None) -> Callable:
    """Decorator that gives the decorated function the image resolution."""

    @functools.wraps(func)
    def wrapper(self: MultiTraceChart, *args, **kwargs):
        if 'width' not in kwargs:
            kwargs['width'] = self.resolution[0]
        if 'height' not in kwargs:
            kwargs['height'] = self.resolution[1]
        if 'resolution' in kwargs:
            kwargs['width'] = kwargs['resolution'][0]
            kwargs['height'] = kwargs['resolution'][1]
            del kwargs['resolution']
        if 'size' in kwargs:
            kwargs['width'] = kwargs['size'][0]
            kwargs['height'] = kwargs['size'][1]
            del kwargs['size']

        return func(self, *args, **kwargs)

    return wrapper


@dataclass(unsafe_hash=True)
class TraceMetadata(FlanaBase):
    """Specifies the settings for a trace of a Plotly chart."""

    type_: plotly.basedatatypes = plotly.graph_objects.Scatter
    name: str = 'trace'
    group: str = None
    legend: str = 'trace'
    show: bool = True
    show_legend: bool = True
    show_markers: bool = False
    show_lines: bool = True
    color: str = 'black'
    opacity: float = 1
    pattern: dict = field(default_factory=dict)
    connect_gaps: bool = True
    default_x_data: float = None
    default_y_data: float = None
    default_min: str | float = 0
    default_max: str | float = 100
    y_tick_suffix: str = ''
    y_delta_tick: float = 10
    hide_y_ticks_if: str = 'False'
    x_data_multiplier: float = 1
    y_data_multiplier: float = 1
    y_axis_width: float = 120


@dataclass(unsafe_hash=True)
class MultiTraceChart(FlanaBase):
    """Class that simplifies the use of Plotly charts with several traces."""

    _font: dict = field(default_factory=dict)
    _legend: dict = field(default_factory=dict)
    _margin: dict = field(default_factory=dict)
    _title: dict = field(default_factory=dict)
    figure: plotly.graph_objects.Figure = field(default_factory=plotly.graph_objects.Figure)
    trace_metadatas: dict[str, TraceMetadata] = ()
    x_data: list = field(default_factory=list)
    all_y_data: list[list] = field(default_factory=list)
    show_middle_horizontal_line: bool = False
    resolution: tuple[int, int] = (1920, 1080)  # 3840×2160, 2560×1440, 2048×1152, 1920×1080, 1366×768, 1280×720, 850×480
    nbinsx: int = None

    def __post_init__(self):
        self.font = self._font
        self.legend = self._legend
        self.margin = self._margin
        self.title = self._title

    def __getattr__(self, item):
        if any(item.startswith(start) for start in ('xaxis', 'yaxis')):
            return vars(self)[item]
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if not any(key.startswith(start) for start in ('xaxis', 'yaxis')):
            super().__setattr__(key, value)
            return

        key_name = key[:5]
        try:
            axis_index = int(key[5:])
        except ValueError:
            axis_index = 1
        key = f'{key_name}{axis_index}'

        try:
            vars(self)[key] |= value
        except KeyError:
            super().__setattr__(key, value)
        self.figure.update_layout({key: value})

    def add_lines(self):
        """Print the x-axis horizontal line."""

        if self.show_middle_horizontal_line:
            self.figure.add_shape(type='line', x0=0, x1=1, y0=0.5, y1=0.5, xref='paper', yref='paper', line_width=1, line_dash='dot')

    def clear(self):
        """Reinitialize the object."""

        self.figure = plotly.graph_objects.Figure()
        for attribute_name in vars(self):
            if any(attribute_name.startswith(start) for start in ('xaxis', 'yaxis')):
                super().__setattr__(attribute_name, {})

        self.__post_init__()

    def draw(self):
        """Apply the trace metadata and the axis data to draw them in the Plotly figure."""

        self.add_lines()

        active_trace_groups = set()
        axis_index = 1
        tick_len = 0
        for trace_metadata, y_data in zip(self.trace_metadatas.values(), self.all_y_data):
            if not trace_metadata.show:
                continue

            show_y_axis = not trace_metadata.group or trace_metadata.group not in active_trace_groups
            active_trace_groups.add(trace_metadata.group)
            self.format_trace(trace_metadata, y_data, axis_index, tick_len, show_y_axis)
            axis_index += 1
            if show_y_axis:
                tick_len += trace_metadata.y_axis_width

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, kwargs: dict):
        self._font |= kwargs
        self.figure.update_layout({'font': self._font})

    def format_trace(self, trace_metadata: TraceMetadata, y_data: Iterable, axis_index: int, tick_len: float, show_y_axis=True):
        """Apply the trace metadata and the axis data to format and draw one trace in the Plotly figure."""

        def adjust_y_axis_height():
            y_data_ = [y_datum for y_datum in y_data if y_datum is not None]

            try:
                min_y_data = min(y_data_)
            except ValueError:
                min_y_data = trace_metadata.default_min
            try:
                max_y_data = max(y_data_)
            except ValueError:
                max_y_data = trace_metadata.default_max

            if isinstance(trace_metadata.default_min, str):
                default_min = float(sympy.sympify(trace_metadata.default_min.format(min_y_data=min_y_data, max_y_data=max_y_data)))
            else:
                default_min = trace_metadata.default_min

            if isinstance(trace_metadata.default_max, str):
                default_max = float(sympy.sympify(trace_metadata.default_max.format(min_y_data=min_y_data, max_y_data=max_y_data)))
            else:
                default_max = trace_metadata.default_max

            min_delta = default_min - min_y_data
            max_delta = max_y_data - default_max
            if min_delta > 0 or max_delta > 0:
                delta = max(min_delta, max_delta)
            else:
                delta = 0
            if delta:
                min_tick = round(default_min - delta - 5, -1)
                max_tick = round(default_max + delta + 5, -1)
            else:
                min_tick = default_min - delta
                max_tick = default_max + delta

            setattr(self, f'yaxis{axis_index}', {'range': (min_tick, max_tick)})

            # ----- Hide ticks -----
            tick_vals = []
            for tick in iterables.frange(min_tick, max_tick, trace_metadata.y_delta_tick, include_last=True):
                if sympy.sympify(trace_metadata.hide_y_ticks_if.format(tick=(tick_ := round(tick, 2)), min_y_data=min_y_data, max_y_data=max_y_data)):
                    tick_vals.append(None)
                else:
                    tick_vals.append(tick_)

            setattr(self, f'yaxis{axis_index}', {'tickvals': tick_vals})

        # ----- Update the data -----
        x_data = self.update_data(self.x_data, trace_metadata.default_x_data, trace_metadata.x_data_multiplier)
        y_data = self.update_data(y_data, trace_metadata.default_y_data, trace_metadata.y_data_multiplier)

        match trace_metadata.show_lines, trace_metadata.show_markers:
            case [True, True]:
                mode = 'lines+markers'
            case [True, False]:
                mode = 'lines'
            case [False, True]:
                mode = 'markers'
            case _:
                mode = None

        if issubclass(trace_metadata.type_, plotly.graph_objects.Scatter):
            extra_trace_kwargs = {
                'mode': mode,
                'line': {'color': trace_metadata.color},
                'connectgaps': trace_metadata.connect_gaps
            }
        else:
            extra_trace_kwargs = {
                'marker': {'color': trace_metadata.color, 'pattern': trace_metadata.pattern},
                'histfunc': 'sum',
                'xbins': {'size': 60 * 60 * 1000}
            }

        trace = trace_metadata.type_(
            x=x_data,
            y=y_data,
            yaxis=f'y{axis_index}',
            name=trace_metadata.legend,
            showlegend=trace_metadata.show_legend,
            opacity=trace_metadata.opacity,
            **extra_trace_kwargs
        )

        # ----- Base kwargs -----
        setattr(self, f'yaxis{axis_index}', {
            'ticks': 'outside',
            'ticklen': tick_len,
            'color': 'rgba(0,0,0,0)',
            'tickfont_color': trace_metadata.color,
            'ticksuffix': trace_metadata.y_tick_suffix,
            'dtick': trace_metadata.y_delta_tick,
            'showticklabels': show_y_axis
        })

        # ----- Overlaying -----
        if axis_index != 1:
            setattr(self, f'yaxis{axis_index}', {'overlaying': 'y'})

        adjust_y_axis_height()

        self.figure.add_trace(trace)

    @property
    def legend(self):
        return self._legend

    @legend.setter
    def legend(self, kwargs: dict):
        self._legend |= kwargs
        self.figure.update_layout({'legend': self._legend})

    @property
    def margin(self):
        return self._margin

    @margin.setter
    def margin(self, kwargs: dict):
        self._margin |= kwargs
        self.figure.update_layout({'margin': self._margin})

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, kwargs: dict):
        self._title |= kwargs
        self.figure.update_layout({'title': self._title})

    @find_resolution
    def show(self, *args, **kwargs):
        self.figure.show(*args, **kwargs)

    @find_resolution
    def to_image(self, *args, **kwargs) -> bytes:
        return self.figure.to_image(*args, **kwargs)

    @staticmethod
    def update_data(data: Iterable, default_value: float, multiplier: float) -> list:
        """Apply a multiplier or default value to each element of the iterable."""

        updated_data = []
        for datum in data:
            if datum is None:
                updated_data.append(default_value)
            else:
                try:
                    updated_data.append(datum * multiplier)
                except TypeError:
                    updated_data.append(datum)

        return updated_data


@dataclass(unsafe_hash=True)
class DateChart(MultiTraceChart):
    """Inherits MultiTraceChart to provide date-dependent state."""

    show_now_vertical_line: bool = True



from __future__ import annotations  # todo0 remove in 3.11

import base64
import binascii
import copy
import datetime
import json
import pickle
import pprint
import typing
from dataclasses import dataclass, field
from enum import Enum
from types import NoneType
from typing import AbstractSet, Any, Iterable, Iterator, Sequence

import pymongo.collection
import pymongo.database
from bson import ObjectId

from flanautils import iterables


class BytesBase:
    """Base class for serialize objects to bytes with pickle."""

    def __bytes__(self):
        return pickle.dumps(self)

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> Any:
        return pickle.loads(bytes_)

    def to_bytes(self):
        return bytes(self)


class CopyBase:
    """Base class for copy and deepcopy objects."""

    def copy(self) -> CopyBase:
        return copy.copy(self)

    def deep_copy(self) -> CopyBase:
        return copy.deepcopy(self)


class DictBase:
    """Base class for serialize objects to dict."""

    def _dict_repr(self) -> Any:
        """
        Returns the dict representation of your object.

        It is the only method that you would want to redefine in most cases to represent data.
        """

        return vars(self).copy()

    @classmethod
    def from_dict(cls, data: dict, lazy=True) -> DictBase:
        """Classmethod that constructs an object given a dictionary."""

        def decode_dict(cls_, data_: dict) -> Any:
            new_data = copy.copy(data_)
            type_hints = typing.get_type_hints(cls_)
            for k, v in data_.items():
                try:
                    type_ = type_hints[k]
                except (KeyError, TypeError):
                    continue

                type_origin = typing.get_origin(type_)
                type_args = typing.get_args(type_)
                try:
                    value_type = type_args[-1]
                except (IndexError, TypeError):
                    value_type = type

                match v:
                    case dict() if issubclass(type_, DictBase) and lazy:
                        continue
                    case dict(dict_) if issubclass(type_, DictBase) and not lazy:
                        new_data[k] = decode_dict(type_, dict_)
                    case [*_] as list_ if not isinstance(list_, set) and type_origin and type_origin is not typing.Union and issubclass(type_origin, Iterable) and issubclass(value_type, DictBase) and not lazy:
                        new_data[k] = [decode_dict(value_type, dict_) for dict_ in list_]
                    case bytes(bytes_):
                        try:
                            new_data[k] = pickle.loads(bytes_)
                        except (pickle.UnpicklingError, EOFError):
                            pass

            return new_data if issubclass(cls_, dict) else cls_(**new_data)

        return decode_dict(cls, data)

    def to_dict(self, pickle_types: tuple | list = (), recursive=False) -> Any:
        """Returns the representation of the object as a dictionary."""

        def encode_obj(obj_) -> Any:
            match obj_:
                case _ if isinstance(obj_, pickle_types):
                    return pickle.dumps(obj_)
                case MongoBase():
                    return obj_.to_dict(pickle_types)
                case [*_, _] as objs:
                    return [encode_obj(obj) for obj in objs]
                case _:
                    return obj_

        if not isinstance(dict_repr := self._dict_repr(), dict):
            return dict_repr

        # noinspection DuplicatedCode
        self_vars = dict_repr.copy()
        for k, v in self_vars.items():
            self_vars[k] = encode_obj(v)

        return self_vars


class JSONBASE:
    """Base class for serialize objects to json."""

    def _json_repr(self) -> Any:
        """
        Returns the JSON representation of your object.

        It is the only method that you would want to redefine in most cases to represent data.
        """

        return vars(self)

    @classmethod
    def from_json(cls, text: str) -> Any:
        """Classmethod that constructs an object given a JSON string."""

        def decode_str(obj_: Any) -> Any:
            """Inner function to decode JSON strings to anything."""

            if not isinstance(obj_, str):
                return obj_
            try:
                bytes_ = base64.b64decode(obj_.encode(), validate=True)
                decoded_obj = pickle.loads(bytes_)
            except (binascii.Error, pickle.UnpicklingError, EOFError):
                return obj_
            return decoded_obj

        def decode_list(obj_: Any) -> Any:
            """Inner function to decode JSON lists to anything."""

            return [decode_str(e) for e in obj_]

        def decode_dict(cls_: Any, dict_: dict) -> Any:
            """Inner function to decode JSON dictionaries to anything."""

            kwargs = {}
            for k, v in dict_.items():
                k = decode_str(k)
                v = decode_str(v)
                if isinstance(v, dict):
                    try:
                        v = decode_dict(typing.get_type_hints(cls_)[k], v)
                    except KeyError:
                        pass
                kwargs[k] = v
            return cls_(**kwargs)

        if not isinstance(text, str):
            raise TypeError(f'must be str, not {type(text).__name__}')

        obj = json.loads(text)
        if isinstance(obj, str):
            return decode_str(obj)
        elif isinstance(obj, list):
            return decode_list(obj)
        elif isinstance(obj, dict):
            return decode_dict(cls, obj)
        else:
            return obj

    def to_json(self, pickle_types: tuple | list = (Enum,), indent: int = None) -> str:
        """Returns the representation of the object as a JSON string."""

        # noinspection PyProtectedMember,PyUnresolvedReferences
        def json_encoder(obj: Any) -> Any:
            if isinstance(obj, bytes):
                return base64.b64encode(obj).decode()

            match obj:
                case obj if isinstance(obj, AbstractSet):
                    return list(obj)
                case JSONBASE():
                    if isinstance(obj, pickle_types) and not obj._json_repr.__qualname__.startswith(obj.__class__.__name__):
                        return pickle.dumps(obj)
                    else:
                        return obj._json_repr()
                case (datetime.date() | datetime.datetime()) as date_time:
                    return str(date_time)
                case _:
                    try:
                        return json.dumps(obj, indent=indent)
                    except TypeError:
                        return repr(obj)

        return json.dumps(self, default=json_encoder, indent=indent)


class MeanBase:
    """Base class for calulate the mean of objects."""

    @classmethod
    def mean(cls, objects: Sequence, ratios: list[float] = None, attribute_names: Iterable[str] = ()) -> MeanBase:
        """
        Classmethod that builds a new object calculating the mean of the objects in the iterable for the attributes
        specified in attribute_names with the provided ratios.

        When calculating the mean, if an attribute is None, the ratio given for that object is distributed among the
        rest whose attributes with that name contain some value other than None.

        By default, ratios is 1 / n_objects for every object in objects.
        """

        if not objects:
            return cls()

        n_objects = len(objects)
        if not ratios:
            ratios = [1 / n_objects for _ in objects]
        elif len(ratios) != len(objects):
            raise ValueError('Wrong ratios length')

        # ----- updates the ratios depending on empty attributes -----
        attributes_ratios = {}
        attributes_ratios_length = {}
        for attribute_name in attribute_names:
            attributes_ratios[attribute_name] = ratios.copy()
            attributes_ratios_length[attribute_name] = len(attributes_ratios[attribute_name])
            for object_index, object_ in enumerate(objects):
                if not object_ or getattr(object_, attribute_name, None) is None:
                    attributes_ratios_length[attribute_name] -= 1
                    try:
                        ratio_part_to_add = attributes_ratios[attribute_name][object_index] / attributes_ratios_length[attribute_name]
                    except ZeroDivisionError:
                        ratio_part_to_add = 0
                    attributes_ratios[attribute_name][object_index] = 0
                    for ratio_index, _ in enumerate(attributes_ratios[attribute_name]):
                        if attributes_ratios[attribute_name][ratio_index]:
                            attributes_ratios[attribute_name][ratio_index] += ratio_part_to_add

        attribute_values = {}
        timezone: datetime.timezone | None = None
        for attribute_name, attribute_ratios in attributes_ratios.items():
            values = []
            for object_, ratio in zip(objects, attribute_ratios):
                if ratio:
                    attribute = getattr(object_, attribute_name)
                    if attribute_name in ('sunrise', 'sunset'):
                        timezone = attribute.tzinfo
                        attribute = attribute.timestamp()
                    values.append(attribute * ratio)

            if values:
                final_value = sum(values)
                if attribute_name in ('sunrise', 'sunset'):
                    final_value = datetime.datetime.fromtimestamp(final_value, timezone)
                attribute_values[attribute_name] = final_value

        # noinspection PyArgumentList
        return cls(**attribute_values)


class MongoBase(DictBase, BytesBase):
    """
    Base class for mapping objects to mongo documents and vice versa (Object Document Mapper).

    Dataclass compatible.
    """

    _id: ObjectId = None
    _database: pymongo.database.Database = None
    _unique_keys: str | Iterable[str] = ()
    _nullable_unique_keys: str | Iterable[str] = ()
    collection: pymongo.collection.Collection = None

    def __init__(self):
        """Automatically generate an ObjectId and store references to the database and collection."""

        match self._id:
            case str() if self._id:
                super().__setattr__('_id', ObjectId(self._id))
            case ObjectId() as object_id:
                super().__setattr__('_id', object_id)
            case _:
                super().__setattr__('_id', ObjectId())

        if self.collection is not None:
            super().__setattr__('collection', self.collection)
            super().__setattr__('_database', self.collection.database)
        if isinstance(self._unique_keys, str):
            super().__setattr__('_unique_keys', (self._unique_keys,))
        if isinstance(self._nullable_unique_keys, str):
            super().__setattr__('_nullable_unique_keys', (self._nullable_unique_keys,))

        self._create_unique_indices()

    def __post_init__(self):
        MongoBase.__init__(self)

    def __bytes__(self):
        return pickle.dumps(self.to_dict())

    def __eq__(self, other):
        if unique_attributes := self.unique_attributes:
            return isinstance(other, self.__class__) and unique_attributes == other.unique_attributes
        else:
            return isinstance(other, self.__class__) and self._id == other._id

    def __getattribute__(self, attribute_name):
        """
        Advice: Do not redefine this method.

        This very hacky method is called every time a class attribute is accessed. It is very difficult to implement
        since it is called recursively.

        So don't redefine this method if you don't really know how python treats objects internally.
        """

        value = super().__getattribute__(attribute_name)

        try:
            type_ = typing.get_type_hints(super().__getattribute__('__class__'))[attribute_name]
        except KeyError:
            return value

        match value:
            case ObjectId() as object_id if issubclass(type_, MongoBase):
                value = type_.find_one({'_id': object_id})
            case [*_, ObjectId()] as object_ids if type_arg := iterables.find(typing.get_args(type_), MongoBase):
                value = [result for object_id in object_ids if (result := type_arg.find_one({'_id': object_id}))]
            case _:
                return value

        super().__setattr__(attribute_name, value)
        return value

    def __hash__(self):
        if unique_attributes := self.unique_attributes:
            return hash(tuple(unique_attributes.values()))
        else:
            return hash(self._id)

    def _create_unique_indices(self):
        """Create the unique indices in the database based on _unique_keys and _nullable_unique_keys attributes."""

        if not self._unique_keys:
            return

        unique_keys = [(unique_key, pymongo.ASCENDING) for unique_key in self._unique_keys]
        type_filter = {'$type': ['number', 'string', 'object', 'array', 'binData', 'objectId', 'bool', 'date', 'regex',
                                 'javascript', 'regex', 'timestamp', 'minKey', 'maxKey']}
        partial_unique_filter = {nullable_unique_key: type_filter for nullable_unique_key in self._nullable_unique_keys}

        self.collection.create_index(unique_keys, partialFilterExpression=partial_unique_filter, unique=True)

    def _dict_repr(self) -> Any:
        return {k: v for k, v in vars(self).items() if k not in ('_database', '_unique_keys', '_nullable_unique_keys', 'collection')}

    def _json_repr(self) -> Any:
        self_vars = vars(self).copy()
        self_vars['_id'] = repr(self_vars['_id'])

        return {k: v for k, v in self_vars.items() if k not in ('_database', '_unique_keys', '_nullable_unique_keys', 'collection')}

    def _mongo_repr(self) -> Any:
        """Returns the object representation to save in mongo database."""

        return {k: v.value if isinstance(v, Enum) else v for k, v in self._dict_repr().items()}

    def delete(self, cascade=False):
        """
        Delete the object from the database.

        If cascade=True all objects whose classes inherit from MongoBase are also deleted.
        """

        if cascade:
            for referenced_object in self.get_referenced_objects():
                referenced_object.delete(cascade)

        self.collection.delete_one({'_id': self._id})

    @classmethod
    def find(cls, query: dict = None, sort_keys: str | Iterable[str | tuple[str, int]] = (), lazy=False) -> Iterator | list:
        """Query the collection."""

        def find_generator() -> Iterator:
            for document in cursor:
                yield cls.from_dict(document)

        match sort_keys:
            case str():
                sort_keys = ((sort_keys, pymongo.ASCENDING),)
            case [[_, *_], *_]:
                pass
            case [*_]:
                sort_keys = [(sort_key, pymongo.ASCENDING) for sort_key in sort_keys]

        cursor: pymongo.cursor.Cursor = cls.collection.find(query)
        if sort_keys:
            cursor.sort(sort_keys)

        if lazy:
            return find_generator()
        else:
            return [cls.from_dict(document) for document in cursor]

    @classmethod
    def find_one(cls, query: dict = None, sort_keys: str | Iterable[str | tuple[str, int]] = ()) -> MongoBase | None:
        """Query the collection and return the first match."""

        return next(cls.find(query, sort_keys, lazy=True), None)

    def find_in_database_by_id(self, object_id: ObjectId) -> dict | None:
        """Find an object in all database collections by its ObjectId."""

        collections = (self._database[name] for name in self._database.list_collection_names())
        return next((document for collection in collections if (document := collection.find_one({'_id': object_id}))), None)

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> Any:
        return cls.from_dict(super().from_bytes(bytes_))

    def get_referenced_objects(self) -> list[MongoBase]:
        """Returns all referenced objects whose classes inherit from MongoBase."""

        referenced_objects = []
        for k, v in vars(self).items():
            match v:
                case MongoBase() as obj:
                    referenced_objects.append(obj)
                case [*_, MongoBase()] as objs:
                    referenced_objects.extend(obj for obj in objs if isinstance(obj, MongoBase))

        return referenced_objects

    @property
    def object_id(self):
        return self._id

    def pull_from_database(self, overwrite_fields: Iterable[str] = ('_id',), exclude: Iterable[str] = ()):
        """
        Updates the values of the current object with the values of the same object located in the database.

        By default, it updates the ObjectId and the attributes of the current object that contain None. You can force
        write from database with database_priority=True (by default database_priority=False).

        Ignore the attributes specified in exclude.
        """

        unique_attributes = self.unique_attributes

        if any(value is None for value in unique_attributes.values()):
            query = {'_id': self._id}
        else:
            query = {}
            for k, v in unique_attributes.items():
                if isinstance(v, MongoBase):
                    v.pull_from_database(overwrite_fields, exclude)
                    v = v._id
                query[k] = v

        if document := self.collection.find_one(query):
            for database_key, database_value in vars(self.from_dict(document)).items():
                self_value = getattr(self, database_key)
                if (
                        database_key not in exclude
                        and
                        (
                                database_key in overwrite_fields and database_value is not None
                                or
                                self_value is None
                                or
                                isinstance(self_value, Iterable) and not self_value
                        )
                ):
                    super().__setattr__(database_key, database_value)

    def resolve(self):
        """Resolve all the ObjectId references (ObjectId -> MongoBase)."""

        for k in vars(self):
            getattr(self, k)

    def save(
        self,
        pickle_types: tuple | list = (AbstractSet,),
        references=True,
        pull_overwrite_fields: Iterable[str] = ('_id',),
        pull_exclude: Iterable[str] = ()
    ):
        """
        Save (insert or update) the current object in the database.

        If references=True it saves the objects without redundancy (MongoBase -> ObjectId).
        """

        if self.collection is None:
            return

        self.pull_from_database(pull_overwrite_fields, pull_exclude)
        for referenced_object in self.get_referenced_objects():
            referenced_object.save(pickle_types, references, pull_overwrite_fields, pull_exclude)

        data = self.to_mongo(pickle_types)

        if references:
            for k, v in data.items():
                match v:
                    case {'_id': ObjectId() as object_id}:
                        data[k] = object_id
                    case [*_, {'_id': ObjectId()}]:
                        data[k] = [obj_data['_id'] for obj_data in v]

        self.collection.find_one_and_update({'_id': self._id}, {'$set': data}, upsert=True)

    def to_mongo(self, pickle_types: tuple | list = (AbstractSet,)) -> Any:
        """Returns the representation of the object as a mongo compatible dictionary."""

        def encode_obj(obj_) -> Any:
            match obj_:
                case MongoBase() if not isinstance(obj_, pickle_types):
                    return obj_.to_mongo(pickle_types)
                case [*_, _] as objs:
                    return [encode_obj(obj) for obj in objs]
                case _ if not isinstance(obj_, (NoneType, int, float, str, bool, bytes, Sequence, dict, datetime.date, datetime.datetime, ObjectId)):
                    return pickle.dumps(obj_)
                case _:
                    return obj_

        if not isinstance(mongo_repr := self._mongo_repr(), dict):
            return mongo_repr

        # noinspection DuplicatedCode
        self_vars = mongo_repr.copy()
        for k, v in self_vars.items():
            self_vars[k] = encode_obj(v)

        return self_vars

    @property
    def unique_attributes(self):
        """
        Property that returns a dictionary with the name of the attributes that must be unique in the database and their
        values.
        """

        unique_attributes = {}
        for unique_key in self._unique_keys:
            attribute_value = getattr(self, unique_key, None)
            unique_attributes[unique_key] = attribute_value.value if isinstance(attribute_value, Enum) else attribute_value

        return unique_attributes


@dataclass(eq=False)
class DCMongoBase(MongoBase):
    """Base class to be inherited by an unfrozen dataclass."""

    _id: ObjectId = field(kw_only=True, default_factory=ObjectId)


@dataclass(eq=False, frozen=True)
class FrozenDCMongoBase(MongoBase):
    """Base class to be inherited by a frozen dataclass."""

    _id: ObjectId = field(kw_only=True, default_factory=ObjectId)


class REPRBase(DictBase):
    """Base class for a nicer objects representation."""

    def __repr__(self):
        # values = vars(self).values()
        # return f"{self.__class__.__name__}({', '.join(repr(v) for v in values)})"
        return str(self)

    def __str__(self):
        formatted = pprint.pformat(self.to_dict())
        return f'{self.__class__.__name__} {formatted[0]}\n {formatted[1:-1]}\n{formatted[-1]}'  # todo1 someday improve the internal objects appearance


class FlanaBase(REPRBase, JSONBASE, CopyBase, BytesBase):
    """Useful mixin."""


# noinspection PyPropertyDefinition
class FlanaEnum(JSONBASE, DictBase, CopyBase, BytesBase, Enum):
    """Useful mixin for enums."""

    def _dict_repr(self) -> Any:
        return self

    def _json_repr(self) -> Any:
        return bytes(self)

    @classmethod
    @property
    def items(cls) -> list[tuple[str, int]]:
        # noinspection PyTypeChecker
        return [item for item in zip(cls.keys, cls.values)]

    @classmethod
    @property
    def keys(cls) -> list[str]:
        return [element.name for element in cls]

    @classmethod
    @property
    def values(cls) -> list[int]:
        return [element.value for element in cls]



import datetime
import json
import pathlib
import random
import re
import secrets
import string
from typing import Iterable, Type, overload

import jellyfish
import unicodedata

from flanautils import constants, iterables
from flanautils.models.ratio_match import RatioMatch


def cartesian_product_string_matching(a_text: str | Iterable[str], b_text: str | Iterable[str], min_ratio: float = 0) -> dict[dict[str, float]]:
    """
    Compare between all the strings of the first iterable with all of the second (cartesian product) and returns a
    dictionary with the scores.
    """

    a_words = a_text.split() if isinstance(a_text, str) else a_text
    b_words = b_text.split() if isinstance(b_text, str) else b_text
    return {a_word: matches for a_word in a_words if (matches := {b_word: ratio for b_word in b_words if (ratio := jellyfish.jaro_winkler_similarity(a_word, b_word)) >= min_ratio})}


def cast_number(text: str, raise_exception=True) -> int | float | str:
    """
    Try to cast a string to a number.

    If raise_exception=False (True by default), returns the input as it is.
    """

    try:
        return int(text)
    except ValueError:
        try:
            return float(text)
        except ValueError:
            if raise_exception:
                raise
            else:
                return text


def find_coordinates(text: str) -> tuple[float, float] | list[tuple[float, float]] | None:
    """
    Find coordinates in string.

    - If one pair of numbers are matched, returns (latitude, longitude)
    - If more of one pair of numbers are matched, returns [(latitude, longitude), ...]
    - If not match, returns None

    >>> find_coordinates('holaaaa 85.1, 44.3515')
    (85.1, 44.3515)
    >>> find_coordinates('ey que pasa+654.1464;,;+ + +-456.4164hola')
    (654.1464, -456.4164)
    >>> find_coordinates('ey que pasa+654.1464;,;+ + -+456.4164hola    40+5.5')
    [(654.1464, 456.4164), (40.0, 5.5)]
    >>> find_coordinates('ey que pasa')

    """

    results = re.findall(r'[-+\d.]+[,;\s+-]+[-+\d.]+', replace(text, {'-': ' -', '+': ' +'}))

    formatted_results = []
    for result in results:
        try:
            words = re.split(r'[,;\s+]+', result)
            latitude, longitude = iterables.filter(words, condition=lambda e: isinstance(e, int | float), cast_numbers=True)
        except ValueError:
            continue
        formatted_results.append((float(latitude), float(longitude)))

    match formatted_results:
        case [single]:
            return single
        case [_, *_]:
            return formatted_results


def find_jsons(text: str | pathlib.Path) -> list[dict]:
    """Find all well formatted JSONs in a string or in a pathlib.Path and return them in dictionaries."""

    text = read_file(text) or text

    jsons = []
    position = 0
    while (position := text.find('{', position)) != -1:
        result, position = json.JSONDecoder().raw_decode(text, position)
        jsons.append(result)
    return jsons


def find_environment_variables(text: str | pathlib.Path) -> dict:
    """Looks for environment variables in a string or in a pathlib.Path in .env format (key=value)."""

    text = read_file(text) or text
    # noinspection PyTypeChecker
    return dict(line.split('=', maxsplit=1) for original_line in text.splitlines() if '=' in (line := original_line.strip()) and not line.startswith('#'))


def join_last_separator(elements: Iterable, separator: str, last_separator: str, final_char='') -> str:
    """
    Join all the elements in a string, using a separator for all of them except the last one, where it uses
    last_separator.

    You can also add to final char.

    >>> join_last_separator(['Uno', 'dos', 'tres', 'cuatro'], ', ', ' y ', '.')
    'Uno, dos, tres y cuatro.'
    """

    elements = list(elements)
    if not len(elements):
        return ''
    if len(elements) == 1:
        return f'{elements[0]}{final_char}'
    return f'{separator.join(elements[:-1])}{last_separator}{elements[-1]}{final_char}'


def numbers_to_words(number: int, language='es') -> str:
    """
    Convert an integer into its textual representation according to the language.

    >>> numbers_to_words(7)
    'siete'
    >>> numbers_to_words(92)
    'noventa y dos'
    >>> numbers_to_words(15)
    'quince'
    >>> numbers_to_words(16)
    'dieciséis'
    """

    if number >= 0:
        is_positive = True
    else:
        is_positive = False
        number *= -1

    if number not in range(0, 101):
        raise ValueError('value out of range')

    if language == 'es':
        number_word_es = constants.NUMBER_WORDS[language]
        try:
            result = number_word_es[number]
        except KeyError:
            tens, units = divmod(number, 10)
            separator = ''
            if tens == 1:
                tens_word = 'dieci'
            elif tens == 2:
                tens_word = 'veinti'
            elif tens == 10:
                tens_word = 'ciento'
            else:
                separator = ' y '
                tens_word = number_word_es[number - units]
            result = f"{tens_word}{separator}{number_word_es[units]}"
        if is_positive:
            return result
        else:
            return f"{number_word_es['-']} {result}"
    else:
        raise NotImplementedError('not implemented for that language')


@overload
def random_string(min_len: int, max_len: int, letters=True, numbers=True, n_spaces=0) -> str:
    pass


@overload
def random_string(str_len=10, letters=True, numbers=True, n_spaces=0) -> str:
    pass


def random_string(min_len=10, max_len=None, letters=True, numbers=True, n_spaces=0) -> str:
    """
    Generate a random string.

    It can generate random strings of random size within the limits specified by arguments.

    You can specify the content of the strings: letters, numbers and number of spaces.
    """

    if isinstance(min_len, bool):
        letters = min_len
        min_len = None
        if isinstance(max_len, bool):
            numbers = max_len
            max_len = None
    elif isinstance(max_len, bool):
        letters = max_len
        max_len = None
    characters = f"{string.ascii_letters if letters else ''}{string.digits if numbers else ''}{' ' * n_spaces}"

    if max_len is not None:
        if max_len < min_len:
            return ''
        min_len = random.randint(min_len, max_len)

    return ''.join(secrets.choice(characters) for _ in range(min_len))


def read_file(path: str | pathlib.Path) -> str | None:
    """
    Returns the content of the file as text.

    Returns None if path is not a valid path string or pathlib.Path.
    """

    match path:
        case str() | pathlib.Path() as path:
            if pathlib.Path(path).is_file():
                with open(path) as file:
                    return file.read()


def remove_accents(text: str, ignore: Iterable = ('ñ', 'ç')) -> str:
    """
    Encode the string to remove the accents from non ascii characters.

    Ignore the characters contained in ignore argument. By default ignore=('ñ', 'ç').

    >>> remove_accents('aáeèiîoöuuºªçñ')
    'aaeeiioouuçñ'
    >>> remove_accents('Mañana iba a salir pero el otro día iba por la calle y casi me atropella un camión que iba muy rápido.')
    'Mañana iba a salir pero el otro dia iba por la calle y casi me atropella un camion que iba muy rapido.'
    """

    return ''.join(char if char in ignore else unicodedata.normalize('NFD', char).encode('ascii', 'ignore').decode() for char in text)


def remove_symbols(text: str, ignore: Iterable = (), replace_with='') -> str:
    """
    Remove the typical symbols from the string.

    Ignore the characters contained in ignore argument. By default ignore=().

    Symbols are replaced with replace_with argument. By default replace_with=''.

    >>> remove_symbols('¿No vás muy rápido? Yo creo que sí.')
    'No vás muy rápido Yo creo que sí'
    >>> remove_symbols('¿No vás muy rápido? Yo creo que sí.', ignore=('?',), replace_with='8')
    '8No vás muy rápido? Yo creo que sí8'
    """

    return ''.join((char if char in ignore else replace_with) if char in constants.SYMBOLS else char for char in text)


replace_symbols = remove_symbols


def replace(text: str, replacements: dict) -> str:
    """
    Returns a copy of text with the replacements applied.

    >>> replace('abc.-.', {'.': '*', '-': '<===>'})
    'abc*<===>*'
    >>> replace('Hola1 que2 ase3', {'Hola': 'Hi', 'que': 'ase', 'ase': None})
    'Hi1 ase2 3'
    """

    if not replacements:
        return text

    result: list[str] = []

    start = 0
    while start < len(text):
        pattern: list[str] = []

        for char in text[start:]:
            pattern.append(char)
            try:
                new_value = replacements[''.join(pattern)]
            except KeyError:
                pass
            else:
                result.append(new_value or '')
                start += len(pattern)
                break
        else:
            result.append(text[start])
            start += 1

    return ''.join(result)


translate = replace


@overload
def str_to_class(class_names: str | Type, globals_: dict) -> Type:
    pass


@overload
def str_to_class(class_names: Iterable[str | Type], globals_: dict) -> tuple[Type, ...]:
    pass


def str_to_class(class_names: str | Type | Iterable[str | Type], globals_: dict) -> Type | tuple[Type, ...]:
    """Converts the class or classes names into Python Types found in globals_."""

    globals_ = globals_ or {}
    if isinstance(class_names, str):
        return globals_[class_names]
    else:
        return tuple(globals_[class_name] for class_name in class_names)


def sum_numbers_in_text(text: str, language='es') -> int:
    """
    Add all the existing numbers in a text, whether they are in numerical or textual form.

    >>> sum_numbers_in_text('Uno más dos. Y luego cuarenta y cuatro y cuarenta y 2 y 10 y 1 más 7.')
    107
    >>> sum_numbers_in_text('0.1 + uno más dos, 5.2 -1.1 mas 5.3 -5.8 ')
    6.7
    """

    n = 0
    for word in text.split():
        try:
            n += cast_number(word.strip('.'))
        except ValueError:
            pass

    return n + words_to_numbers(text, language=language)


def words_to_numbers(text: str, ignore_no_numbers=True, language='es') -> int:
    """
    Convert all numbers in textual representation, according to the language, into an integer and return the sum of
    them.

    >>> words_to_numbers('Borra veintidos mensajes. Y... luego borra otros treinta y cinco.')
    57
    """

    text = remove_accents(text)

    if language == 'es':
        number_words_es = constants.NUMBER_WORDS[language]
        text = re.sub(r'(([ei]nt[aeio])|(ec))', r'\1 ', text)
        text = replace(text, {symbol: None for symbol in constants.SYMBOLS} | {'y': ' '})
        words = text.lower().split()
        n = 0
        sign_ = 1
        for word in words:
            if jellyfish.jaro_winkler_similarity(word, number_words_es['-']) >= constants.NUMBERS_RATIO_MATCHING:
                sign_ = -1
                continue

            word_possible_matches = []
            for number_word in number_words_es.values():
                ratio = jellyfish.jaro_winkler_similarity(word, number_word)
                if ratio >= constants.NUMBERS_RATIO_MATCHING:
                    word_possible_matches.append(RatioMatch(number_words_es[number_word], ratio))

            if word_possible_matches:
                n += sign_ * max(word_possible_matches, key=lambda match: match.ratio).element
            elif not ignore_no_numbers:
                raise KeyError(word)

        return n

    raise NotImplementedError('not implemented for that language')


def words_to_time(text: str, language='es') -> datetime.timedelta:
    """
    Convert time in textual representation, according to the language, into a datetime.timedelta.

    >>> words_to_time('Un minuto y 10 segundos.')
    datetime.timedelta(seconds=70)
    """

    if language == 'es':
        delta_time = datetime.timedelta()
        n = 0
        for word in text.split():
            if jellyfish.jaro_winkler_similarity(word, 'segundo') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(seconds=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'minuto') >= constants.TIME_UNITS_RATIO_MATCHING or jellyfish.jaro_winkler_similarity(word, 'min') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(minutes=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'hora') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(hours=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'dia') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(days=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'semana') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(weeks=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'mes') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(weeks=n * 4.34524)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'año') >= constants.TIME_UNITS_RATIO_MATCHING:
                delta_time += datetime.timedelta(weeks=n * 52.1429)
                n = 0
            else:
                n += sum_numbers_in_text(word)

        return delta_time

    raise NotImplementedError('not implemented for that language')



def sign(number: int | float) -> int:
    """
    Return the number sign.

    >>> sign(5)
    1
    >>> sign(-5.2)
    -1
    >>> sign(0)
    0
    """

    if number > 0:
        return 1
    elif number < 0:
        return -1
    return 0



from enum import auto

from flanautils.models.bases import FlanaEnum


class MediaType(FlanaEnum):
    AUDIO = auto()
    ERROR = auto()
    GIF = auto()
    IMAGE = auto()
    VIDEO = auto()

    @property
    def extension(self) -> str:
        if self is MediaType.AUDIO:
            return 'mp3'
        elif self is MediaType.GIF:
            return 'gif'
        elif self is MediaType.IMAGE:
            return 'jpg'
        elif self is MediaType.VIDEO:
            return 'mp4'
        else:
            return ''


class Source(FlanaEnum):
    INSTAGRAM = auto()
    LOCAL = auto()
    TIKTOK = auto()
    TWITTER = auto()
    REDDIT = auto()
    YOUTUBE = auto()

    @property
    def name(self):
        if self is Source.INSTAGRAM:
            return 'Instagram'
        elif self is Source.TIKTOK:
            return 'TikTok'
        elif self is Source.TWITTER:
            return 'Twitter'
        elif self is Source.REDDIT:
            return 'Reddit'
        elif self is Source.YOUTUBE:
            return 'YouTube'



from dataclasses import dataclass

from flanautils import constants, strings
from flanautils.models.bases import FlanaBase


@dataclass(unsafe_hash=True)
class TimeUnits(FlanaBase):
    """
    Represents the time information grouping it into typical units.

    You can represent that tense in textual form according to the language.

    >>> time_units = TimeUnits(hours=1.5, seconds=130)
    >>> time_units
    TimeUnits(years=0, months=0, weeks=0, days=0, hours=1, minutes=32, seconds=10)
    >>> time_units.to_words()
    '1 hora, 32 minutos y 10 segundos'
    >>> time_units = TimeUnits(years=1/12, days=1, minutes=0.1)
    >>> time_units
    TimeUnits(years=0, months=1, weeks=0, days=1, hours=0, minutes=0, seconds=6)
    >>> time_units.to_words()
    '1 mes, 1 día y 6 segundos'
    >>> time_units = TimeUnits(days=180)
    >>> time_units
    TimeUnits(years=0, months=5, weeks=3, days=6, hours=21, minutes=59, seconds=54)
    >>> time_units.to_words()
    '5 meses, 3 semanas, 6 días, 21 horas, 59 minutos y 54 segundos'
    """

    years: int = 0
    months: int = 0
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0

    def __init__(self, years: float = 0, months: float = 0, weeks: float = 0, days: float = 0, hours: float = 0, minutes: float = 0, seconds: float = 0):
        quotient, remainder = divmod(seconds, 60)
        seconds = remainder
        minutes += quotient

        quotient, remainder = divmod(minutes, 60)
        minutes = remainder
        hours += quotient

        quotient, remainder = divmod(hours, 24)
        hours = remainder
        days += quotient

        quotient, remainder = divmod(days, 7)
        days = remainder
        weeks += quotient

        quotient, remainder = divmod(weeks, constants.WEEKS_IN_A_MONTH)
        weeks = remainder
        months += quotient

        quotient, remainder = divmod(months, 12)
        months = remainder
        years += int(quotient)

        months += years % 1 * 12
        weeks += months % 1 * constants.WEEKS_IN_A_MONTH
        days += weeks % 1 * 7
        hours += days % 1 * 24
        minutes += hours % 1 * 60
        seconds += minutes % 1 * 60

        self.years = int(years)
        self.months = int(months)
        self.weeks = int(weeks)
        self.days = int(days)
        self.hours = int(hours)
        self.minutes = int(minutes)
        self.seconds = int(seconds)

    def to_words(self, language='es') -> str:
        if language == 'es':
            translation = {
                'years': {'singular': 'año', 'plural': 'años'},
                'months': {'singular': 'mes', 'plural': 'meses'},
                'weeks': {'singular': 'semana', 'plural': 'semanas'},
                'days': {'singular': 'día', 'plural': 'días'},
                'hours': {'singular': 'hora', 'plural': 'horas'},
                'minutes': {'singular': 'minuto', 'plural': 'minutos'},
                'seconds': {'singular': 'segundo', 'plural': 'segundos'}
            }
            separator = ', '
            last_separator = ' y '
        elif language == 'en':
            translation = {
                'years': {'singular': 'year', 'plural': 'years'},
                'months': {'singular': 'month', 'plural': 'months'},
                'weeks': {'singular': 'week', 'plural': 'weeks'},
                'days': {'singular': 'day', 'plural': 'days'},
                'hours': {'singular': 'hour', 'plural': 'hours'},
                'minutes': {'singular': 'minute', 'plural': 'minutes'},
                'seconds': {'singular': 'second', 'plural': 'seconds'}
            }
            separator = ', '
            last_separator = ' and '
        else:
            raise NotImplementedError('not implemented for that language')

        words = (f"{v} {translation[k]['singular'] if v == 1 else translation[k]['plural']}" for k, v in vars(self).items() if v)
        return strings.join_last_separator(words, separator, last_separator)

    def total_seconds(self) -> int:
        months = 12 * self.years
        weeks = constants.WEEKS_IN_A_MONTH * (self.months + months)
        days = 7 * (self.weeks + weeks)
        hours = 24 * (self.days + days)
        minutes = 60 * (self.hours + hours)
        seconds = 60 * (self.minutes + minutes)

        return self.seconds + seconds



from __future__ import annotations  # todo0 remove in 3.11

import itertools
import json
import math
import pickle
from typing import AbstractSet, Any, Generic, Iterable, Iterator, MutableSet, Type, TypeVar

from flanautils import iterables
from flanautils.models.bases import FlanaBase, JSONBASE

E = TypeVar('E')


class OrderedSet(FlanaBase, MutableSet, Generic[E]):
    """
    Set that maintains the insertion order.

    Reimplements all set and list methods.

    Note that being built on hashes, the list methods are not efficient since the sets are not designed to access a
    specific element of the structure.
    """

    FLATTEN_DEPTH = 1

    T = TypeVar('T', bound='OrdereSet')

    def __init__(self, *args: Any):
        self._elements_dict = {element: None for element in iterables.flatten_iterator(*args, depth=self.FLATTEN_DEPTH)}

    def __add__(self, other) -> OrderedSet[E]:
        return self | other

    def __iadd__(self, other):
        return self.__ior__(other)

    def __radd__(self, other) -> OrderedSet[E]:
        return other | self

    def __and__(self, other) -> OrderedSet[E]:
        return OrderedSet(value for value in self if value in self.ordered_set_if_not_set(other))

    def __iand__(self, other):
        for value in (self - self.ordered_set_if_not_set(other)):
            self.discard(value)
        return self

    def __rand__(self, other) -> OrderedSet[E]:
        return self.ordered_set_if_not_set(other) & self

    def __contains__(self, element: Any) -> bool:
        return element in self._elements_dict

    def __delitem__(self, item):
        if isinstance(item, slice):
            self.discard_many(self[item])
        else:
            self.discard(self[item])

    def __eq__(self, other) -> bool:
        if not isinstance(other, OrderedSet):
            return False

        for self_element, other_element in itertools.zip_longest(self, other):
            if self_element != other_element:
                return False
        return True

    def __getitem__(self, item) -> E | OrderedSet[E]:
        ordered_set = OrderedSet()
        len_self = len(self)

        if isinstance(item, slice):
            step = 1 if item.step is None else item.step
            if step > 0:
                if item.start is None:
                    start = 0
                else:
                    start = self.positive_index(item.start)
                    if start >= len_self:
                        return ordered_set
                    start = max(0, start)

                if item.stop is None:
                    stop = len_self
                else:
                    stop = self.positive_index(item.stop)
                    if stop <= 0:
                        return ordered_set
                    stop = min(stop, len_self)

                iterator = iter(self)

            elif step < 0:
                if item.start is None:
                    start = 0
                else:
                    start = self.positive_index(item.start)
                    if start < 0:
                        return ordered_set
                    start = len_self - 1 - min(start, len_self - 1)

                if item.stop is None:
                    stop = len_self
                else:
                    stop = self.positive_index(item.stop)
                    if stop >= len_self - 1:
                        return ordered_set
                    stop = len_self - 1 - max(-1, stop)

                step *= -1

                iterator = reversed(self)

            else:
                raise ValueError('slice step cannot be zero')

            iterables.iterate_n(iterator, start)
            for _ in range(math.ceil((stop - start) / step)):
                try:
                    ordered_set.add(next(iterator))
                    iterables.iterate_n(iterator, step - 1)
                except StopIteration:
                    break

            return ordered_set

        elif isinstance(item, int):
            item = self.positive_index(item)
            if not 0 <= item < len_self:
                raise IndexError('index out of range')

            if item <= len_self / 2:
                iterator = iter(self)
                index = item
            else:
                iterator = reversed(self)
                index = len_self - item - 1

            return iterables.iterate_n(iterator, index + 1)

        else:
            raise TypeError('indices must be integers or slices')

    def __iter__(self) -> Iterator[E]:
        return iter(self._elements_dict)

    def __len__(self) -> int:
        return len(self._elements_dict)

    def __or__(self, other) -> OrderedSet[E]:
        return OrderedSet(super().__or__(self.ordered_set_if_not_set(other)))

    def __ior__(self, other):
        for value in self.ordered_set_if_not_set(other):
            self.add(value)
        return self

    def __ror__(self, other) -> OrderedSet[E]:
        return self.ordered_set_if_not_set(other) | self

    def __repr__(self) -> str:
        return str(self)

    def __reversed__(self) -> Iterator[E]:
        return reversed(self._elements_dict)

    def __str__(self) -> str:
        return f"#{{{', '.join(repr(element) for element in self)}}}"

    def __sub__(self, other) -> OrderedSet[E]:
        return OrderedSet(super().__sub__(self.ordered_set_if_not_set(other)))

    def __isub__(self, other):
        if other is self:
            self.clear()
        else:
            for value in self.ordered_set_if_not_set(other):
                self.discard(value)
        return self

    def __rsub__(self, other) -> OrderedSet[E]:
        return self.ordered_set_if_not_set(other) - self

    def _json_repr(self) -> Any:
        return [json.loads(element.to_json()) if isinstance(element, JSONBASE) else pickle.dumps(element) for element in self]

    def positive_index(self, index_: int) -> int:
        if index_ < 0:
            index_ += len(self)
        return index_

    @classmethod
    def ordered_set_if_not_set(cls: Type[T], arg: Any) -> AbstractSet | T:
        return arg if isinstance(arg, AbstractSet) else cls(iterables.flatten_iterator(arg))

    def add(self, element: Any):
        self._elements_dict[element] = None

    def add_many(self, elements: Iterable):
        for element in elements:
            self.add(element)

    def clear(self):
        try:
            while True:
                self.pop()
        except IndexError:
            pass

    def copy(self) -> OrderedSet[E]:
        return OrderedSet(self)

    def difference(self, *args: Iterable) -> OrderedSet[E]:
        new_ordered_set = OrderedSet(self)
        new_ordered_set.difference_update(*args)
        return new_ordered_set

    def difference_update(self, *args: Iterable):
        self.discard_many(iterables.flatten_iterator(*args, depth=self.FLATTEN_DEPTH))

    def discard(self, element: Any):
        try:
            self._elements_dict.pop(element, None)
        except TypeError:
            pass

    def discard_many(self, elements: Iterable):
        for element in elements:
            self.discard(element)

    def index(self, element, start=None, stop=None, raise_exception=False) -> int:
        if start is not None or stop is not None:
            ordered_set = self[start:stop]
        else:
            ordered_set = self

        if start is None:
            start = 0
        else:
            start = max(0, start)

        for i, element_ in enumerate(ordered_set):
            if element == element_:
                return start + i

        if raise_exception:
            raise ValueError(f'{element} is not in the OrderedSet')

    def insert(self, i, element):
        if element in self:
            return

        i = self.positive_index(i)

        deleted_elements = []
        for i_, element_ in enumerate(list(self)):
            if i <= i_:
                self.discard(element_)
                deleted_elements.append(element_)

        self.add(element)
        self.add_many(deleted_elements)

    def intersection(self, *args: Iterable) -> OrderedSet[E]:
        new_ordered_set = OrderedSet(self)
        new_ordered_set.intersection_update(*args)
        return new_ordered_set

    def intersection_update(self, *args: Iterable):
        elements_to_delete = []
        for element in self:
            for arg in args:
                if element not in iterables.flatten_iterator(arg, depth=self.FLATTEN_DEPTH):
                    elements_to_delete.append(element)
                    break

        self.discard_many(elements_to_delete)

    def is_disjoint(self, other) -> bool:
        return not (self & other)

    def is_subset(self, other) -> bool:
        return self <= other

    def is_superset(self, other) -> bool:
        return self >= other

    def pop(self, index=-1) -> E:
        element = self[index]
        self.discard(element)
        return element

    def reverse(self):
        reversed_self = list(reversed(self))
        self.clear()
        self.add_many(reversed_self)

    def sort(self, key=None, reverse=False):
        sorted_values = sorted(self, key=key, reverse=reverse)
        self.clear()
        self.add_many(sorted_values)

    def symmetric_difference(self, *args: Iterable) -> OrderedSet[E]:
        new_ordered_set = OrderedSet(self)
        new_ordered_set.symmetric_difference_update(*args)
        return new_ordered_set

    def symmetric_difference_update(self, *args: Iterable):
        for arg in args:
            other_minus_self = self.ordered_set_if_not_set(arg) - self
            self.difference_update(arg)
            self.update(other_minus_self)

    def union(self, *args: Iterable) -> OrderedSet[E]:
        new_ordered_set = OrderedSet(self)
        new_ordered_set.update(*args)
        return new_ordered_set

    def update(self, *args: Iterable):
        self.add_many(iterables.flatten_iterator(*args, depth=self.FLATTEN_DEPTH))

    union_update = update



import contextlib
import os
import pathlib
import pkgutil
import subprocess
from contextlib import contextmanager

from flanautils import strings


def resolve_path(path: str) -> pathlib.Path:
    """Resolves actual file or directory paths in distributed libraries."""

    path = path.strip('/')
    try:
        path, file = path.rsplit('/', 1)
    except ValueError:
        file = ''
    path = path.replace('/', '.')

    return pathlib.Path(pkgutil.resolve_name(path).__path__[0]) / file


def set_windows_environment_variables(variables: str | dict | pathlib.Path, search_jsons=True, set_in_system=False):
    """
    Set environment variables for Windows.

    If variables is a str or a pathlib.Path it looks for environment variables in that file in .env or json format.
    """

    match variables:
        case str() | pathlib.Path() as text:
            if pathlib.Path(text).is_file():
                with open(text) as file:
                    text = file.read()

            variables = {}
            if search_jsons:
                variables = {k: v for dict_ in strings.find_jsons(text) for k, v in dict_.items()}
            if not variables:
                variables = strings.find_environment_variables(text)
        case dict(variables):
            pass
        case _:
            raise TypeError('bad arguments')

    if set_in_system:
        for k, v in variables.items():
            subprocess.run(f'setx /m {k} {v}')
    else:
        for k, v in variables.items():
            os.environ[k] = v


@contextmanager
def suppress_stderr():
    """A context manager that redirects stderr to devnull."""

    with open(os.devnull, 'w') as fnull:
        with contextlib.redirect_stderr(fnull) as err:
            yield err


@contextmanager
def suppress_stdout():
    """A context manager that redirects stdout to devnull."""

    with open(os.devnull, 'w') as fnull:
        with contextlib.redirect_stdout(fnull) as out:
            yield out



import datetime
import random
from typing import Iterable, Sequence

from flanautils.data_structures.bi_dict import BiDict
from flanautils.data_structures.ordered_set import OrderedSet

NUMBER_WORDS = {
    'es': BiDict({
        '-': 'menos',
        0: 'cero', 1: 'uno', 2: 'dos', 3: 'tres', 4: 'cuatro', 5: 'cinco', 6: 'seis', 7: 'siete', 8: 'ocho', 9: 'nueve',
        10: 'diez', 11: 'once', 12: 'doce', 13: 'trece', 14: 'catorce', 15: 'quince', 16: 'dieciséis',
        20: 'veinte', 30: 'treinta', 40: 'cuarenta', 50: 'cincuenta', 60: 'sesenta', 70: 'setenta', 80: 'ochenta', 90: 'noventa',
        100: 'cien'
    })
}
NUMBERS_RATIO_MATCHING = 0.9
TIME_UNITS_RATIO_MATCHING = 0.9
WEEKS_IN_A_MONTH = 4.34524
SYMBOLS = ('!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@',
           '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~', '¡', '¨', 'ª', '¬', '´', '·', 'º', '¿', '€')


# noinspection PyPropertyDefinition
class CommonWords:
    common_words = {
        'en': {
            'adverbs': (
                'abnormally', 'above', 'abroad', 'absentmindedly', 'accidentally', 'actually', 'adventurously',
                'afterwards', 'almost', 'already', 'always', 'annually', 'anxiously', 'anywhere', 'arrogantly', 'away',
                'awkwardly', 'back', 'badly', 'barely', 'bashfully', 'beautifully', 'behind', 'bitterly', 'bleakly',
                'blindly', 'blissfully', 'boastfully', 'boldly', 'bravely', 'briefly', 'brightly', 'briskly', 'broadly',
                'busily', 'calmly', 'carefully', 'carelessly', 'cautiously', 'certainly', 'cheerfully', 'clearly',
                'cleverly', 'closely', 'coaxingly', 'colorfully', 'commonly', 'completely', 'continually', 'coolly',
                'correctly', 'courageously', 'crossly', 'cruelly', 'curiously', 'daily', 'daintily', 'deal', 'dearly',
                'deceivingly', 'deeply', 'defiantly', 'deliberately', 'delightfully', 'diligently', 'dimly',
                'doubtfully', 'downstairs', 'dreamily', 'eagerly', 'easily', 'elegantly', 'elsewhere', 'emotely',
                'energetically', 'enormously', 'enough', 'enthusiastically', 'entirely', 'equally', 'especially',
                'even', 'evenly', 'eventually', 'ever', 'exactly', 'excitedly', 'extremely', 'fairly', 'faithfully',
                'famously', 'far', 'fast', 'fatally', 'ferociously', 'fervently', 'few', 'fiercely', 'fondly',
                'foolishly', 'fortunately', 'frankly', 'frantically', 'freely', 'frenetically', 'frequently',
                'frightfully', 'fully', 'furiously', 'generally', 'generously', 'gently', 'gladly', 'gleefully', 'good',
                'gracefully', 'gratefully', 'greatly', 'greedily', 'grimly', 'happily', 'hardly', 'hastily',
                'healthily', 'heavily', 'helpfully', 'helplessly', 'here', 'highly', 'honestly', 'hopelessly', 'hourly',
                'hungrily', 'immediately', 'indoor', 'innocently', 'inquisitively', 'inside', 'instantly', 'intensely',
                'intently', 'interestingly', 'inwardly', 'irritably', 'jaggedly', 'jealously', 'jovially', 'joyfully',
                'joyously', 'jubilantly', 'judgmentally', 'just', 'justly', 'keenly', 'kiddingly', 'kindheartedly',
                'kindly', 'knavishly', 'knowingly', 'knowledgeably', 'kookily', 'last', 'later', 'lazily', 'less',
                'lightly', 'likely', 'limply', 'little', 'lively', 'loftily', 'longingly', 'loosely', 'lot', 'lots',
                'loudly', 'lovingly', 'lowly', 'loyally', 'madly', 'majestically', 'many', 'meaningfully',
                'mechanically', 'merrily', 'miserably', 'mockingly', 'month', 'monthly', 'more', 'mortally', 'mostly',
                'much', 'mysteriously', 'naturally', 'nearby', 'nearly', 'neatly', 'nervously', 'never', 'nicely',
                'noisily', 'normally', 'not', 'nothing', 'now', 'obediently', 'obnoxiously', 'occasionally', 'oddly',
                'offensively', 'officially', 'often', 'only', 'openly', 'optimistically', 'out', 'outside',
                'overconfidently', 'painfully', 'partially', 'patiently', 'perfectly', 'physically', 'playfully',
                'politely', 'poorly', 'positively', 'potentially', 'powerfully', 'promptly', 'properly', 'punctually',
                'quaintly', 'queasily', 'queerly', 'questionably', 'quicker', 'quickly', 'quietly', 'quirkily', 'quite',
                'quizzically', 'randomly', 'rapidly', 'rarely', 'rather', 'readily', 'really', 'reassuringly',
                'recklessly', 'regularly', 'reluctantly', 'repeatedly', 'reproachfully', 'restfully', 'righteously',
                'rightfully', 'rigidly', 'roughly', 'rudely', 'safely', 'scarcely', 'scarily', 'searchingly',
                'secretly', 'sedately', 'seemingly', 'seldom', 'selfishly', 'separately', 'seriously', 'shakily',
                'sharply', 'sheepishly', 'shrilly', 'shyly', 'silently', 'sleepily', 'slowly', 'smoothly', 'softly',
                'solemnly', 'solidly', 'some', 'sometimes', 'soon', 'specifically', 'speedily', 'stealthily', 'sternly',
                'strictly', 'strongly', 'successfully', 'suddenly', 'supposedly', 'surprisingly', 'suspiciously',
                'sweetly', 'swiftly', 'sympathetically', 'tenderly', 'tensely', 'terribly', 'thankfully', 'then',
                'there', 'thoroughly', 'thoughtfully', 'tightly', 'today', 'tomorrow', 'tonight', 'too', 'towards',
                'tremendously', 'triumphantly', 'truly', 'truthfully', 'ultimately', 'unabashedly', 'unaccountably',
                'unbearably', 'unethically', 'unexpectedly', 'unfortunately', 'unimpressively', 'unnaturally',
                'unnecessarily', 'upbeat', 'upright', 'upside-down', 'upward', 'urgently', 'usefully', 'uselessly',
                'usually', 'utterly', 'vacantly', 'vaguely', 'vainly', 'valiantly', 'vastly', 'verbally', 'very',
                'viciously', 'victoriously', 'violently', 'vivaciously', 'voluntarily', 'warmly', 'weakly', 'wearily',
                'week', 'weetly', 'well', 'wetly', 'wholly', 'wildly', 'willfully', 'wisely', 'wishfully', 'woefully',
                'wonderfully', 'worriedly', 'wrongly', 'yawningly', 'year', 'yearly', 'yearningly', 'yesterday',
                'yieldingly', 'youthfully', 'zealously', 'zestfully', 'zestily'),
            'conjunctions': ('and', 'or'),
            'determiners': {
                'definite_articles': ('the',),
                'indefinites': ('another', 'any', 'every', 'few', 'lot', 'much', 'other', 'several', 'some', 'very'),
                'indefinite_articles': ('a', 'an'),
                'possessives': ('hers', 'his', 'its', 'my', 'our', 'their', 'whose', 'your'),
                'demonstratives': ('that', 'these', 'this', 'those'),
                'numerals': (),
                'ordinals': ('first', 'second', 'third', 'fourth', 'fifth', 'sixth', 'seventh', 'eighth', 'ninth',
                             'tenth'),
            },
            'greetings': ('Good morning.', 'Good afternoon.', 'Goodnight.'),
            'prepositions': ('about', 'above', 'abroad', 'accordance', 'according', 'account', 'across', 'addition',
                             'after', 'against', 'ago', 'ahead', 'along', 'amidst', 'among', 'amongst', 'apart',
                             'around', 'as', 'aside', 'at', 'away', 'because', 'before', 'behalf', 'behind', 'below',
                             'beneath', 'beside', 'besides', 'between', 'beyond', 'but', 'by', 'case', 'close',
                             'despite', 'down', 'due', 'during', 'except', 'farwell', 'for', 'from', 'front', 'hence',
                             'in', 'inside', 'instead', 'into', 'lieu', 'like', 'means', 'near', 'next',
                             'notwithstanding', 'of', 'off', 'on', 'onto', 'opposite', 'out', 'outside', 'over',
                             'owing', 'past', 'per', 'place', 'prior', 'round', 'since', 'spite', 'than', 'through',
                             'throughout', 'till', 'times', 'to', 'top', 'toward', 'towards', 'under', 'underneath',
                             'unlike', 'until', 'unto', 'up', 'upon', 'via', 'view', 'with', 'within', 'without',
                             'worth'),
            'pronouns': {
                'interrogatives': ('what', 'which', 'who', 'whom', 'whose'),
                'objects': ('her', 'him', 'it', 'me', 'them', 'us', 'you'),
                'possessives': ('hers', 'his', 'its', 'mine', 'ours', 'theirs', 'yours'),
                'reflexives': ('herself', 'himself', 'itself', 'myself', 'ourselves', 'themselves', 'yourself',
                               'yourselves'),
                'subjects': ('he', 'i', 'it', 'she', 'they', 'we', 'you'),
            },
            'others': ("don't", 'dont', 'let', "let's", 'lets', 'no', 'not', 'yes')
        },
        'es': {
            'adverbs': (
                'abajo', 'aca', 'acaso', 'actualmente', 'ademas', 'ahi', 'ahora', 'algo', 'alla', 'alli', 'alrededor',
                'anoche', 'anteriormente', 'antes', 'aprisa', 'aproximadamente', 'aqui', 'arriba', 'asi', 'asimismo',
                'atras', 'aun', 'ayer', 'bastante', 'bien', 'casi', 'cerca', 'ciertamente', 'cierto', 'claro', 'como',
                'concretamente', 'debajo', 'delante', 'demasiado', 'dentro', 'deprisa', 'despacio', 'despues', 'detras',
                'efectivamente', 'encima', 'enfrente', 'enseguida', 'estupendamente', 'exacto', 'facilmente',
                'fielmente', 'fuera', 'hoy', 'inclusive', 'incluso', 'jamas', 'lejos', 'luego', 'mal', 'mas', 'mañana',
                'medio', 'mejor', 'menos', 'mientras', 'mismamente', 'mucho', 'muy', 'nada', 'negativamente', 'no',
                'nunca', 'obvio', 'peor', 'poco', 'posiblemente', 'precisamente', 'primeramente', 'probablemente',
                'prontamente', 'pronto', 'propiamente', 'proximamente', 'quiza', 'quizas', 'regular',
                'responsablemente', 'seguramente', 'si', 'siempre', 'siquiera', 'solamente', 'solo', 'tal', 'tambien',
                'tampoco', 'tan', 'tanto', 'tarde', 'temprano', 'todavia', 'todo', 'ultimamente', 'unicamente',
                'verdaderamente', 'vez', 'viceversa', 'ya'),
            'conjunctions': ('o', 'y'),
            'determiners': {
                'definite_articles': ('al', 'del', 'el', 'la', 'las', 'lo', 'los'),
                'indefinites': ('algun', 'alguna', 'algunas', 'alguno', 'algunos', 'bastante', 'bastantes',
                                'cualesquiera', 'cualquier', 'cualquiera', 'demasiada', 'demasiadas', 'demasiado',
                                'demasiados', 'misma', 'mismas', 'mismo', 'mismos', 'mucha', 'muchas', 'mucho',
                                'muchos', 'ningun', 'ninguna', 'ningunas', 'ninguno', 'ningunos', 'otra', 'otras',
                                'otro', 'otros', 'poca', 'pocas', 'poco', 'pocos', 'tal', 'tales', 'tanta', 'tantas',
                                'tanto', 'tantos', 'toda', 'todas', 'todo', 'todos', 'uno', 'varias', 'varios'),
                'indefinite_articles': ('un', 'una', 'unas', 'unos'),
                'possessives': ('mi', 'mia', 'mias', 'mio', 'mios', 'mis', 'nuestra', 'nuestras', 'nuestro', 'nuestros',
                                'su', 'sus', 'suya', 'suyas', 'suyo', 'suyos', 'tu', 'tus', 'tuya', 'tuyas', 'tuyo',
                                'tuyos', 'vuestra', 'vuestras', 'vuestro', 'vuestros'),
                'demonstratives': ('aquel', 'aquella', 'aquellas', 'aquello', 'aquellos', 'esa', 'esas', 'ese', 'eso',
                                   'esos', 'esta', 'estas', 'este', 'estos'),
                'numerals': (),
                'ordinals': ('primero', 'segundo', 'tercero', 'cuarto', 'quinto', 'sexto', 'septimo', 'octavo',
                             'noveno', 'decimo'),
            },
            'greetings': ('Buenos días.', 'Buenas tardes.', 'Buenas noches.'),
            'prepositions': ('a', 'ante', 'bajo', 'cabe', 'con', 'contra', 'de', 'desde', 'durante', 'en', 'entre',
                             'hacia', 'hasta', 'mediante', 'para', 'por', 'según', 'sin', 'so', 'sobre', 'tras',
                             'versus', 'vía'),
            'pronouns': {
                'interrogatives': ('cual', 'cuales', 'cuanta', 'cuantas', 'cuanto', 'cuantos', 'que'),
                'objects': ('conmigo', 'consigo', 'contigo', 'el', 'ella', 'ellas', 'ello', 'ellos', 'mi', 'nosotras',
                            'nostros', 'si', 'ti', 'usted', 'ustedes', 'vos', 'vosotras', 'vosotros'),
                'possessives': ('mia', 'mias', 'mio', 'mios', 'nuestra', 'nuestras', 'nuestro', 'nuestros', 'suya',
                                'suyas', 'suyo', 'suyos', 'tuya', 'tuyas', 'tuyo', 'tuyos', 'vuestra', 'vuestras',
                                'vuestro', 'vuestros'),
                'reflexives': ('me', 'nos', 'os', 'se', 'te'),
                'subjects': ('el', 'ella', 'ellas', 'ello', 'ellos', 'la', 'las', 'le', 'les', 'lo', 'los', 'me', 'nos',
                             'nosotras', 'nosotros', 'os', 'se', 'te', 'tu', 'usted', 'ustedes', 'vos', 'vosotras',
                             'vosotros', 'yo'),
            },
            'others': ('ha', 'hace', 'hacer', 'hara', 'he', 'no', 'si', 'va', 'vamos', 'voy')
        }
    }

    @classmethod
    def _get_language_names(cls, languages: str | Sequence[str] = ()) -> Iterable[str]:
        match languages:
            case _ if not languages:
                return cls.common_words.keys()
            case str():
                return languages,
            case [*_]:
                return languages

    @classmethod
    def get(cls, keys: Iterable[str] = (), languages: str | Sequence[str] = ()) -> list[str]:
        def recursive_get(data: Iterable | dict, return_sequence=False) -> Sequence:
            match data:
                case [*_]:
                    if return_sequence:
                        return data
                    else:
                        return ()
                case dict():
                    words_ = OrderedSet()
                    for k, v in data.items():
                        words_ += recursive_get(v, return_sequence or k in keys)
                    return words_

        words = OrderedSet()

        for language_name in cls._get_language_names(languages):
            words += recursive_get(cls.common_words[language_name], not keys)

        return list(words)

    @classmethod
    def random_time_greeting(cls, language: str = None) -> str:
        language = language or random.choice(list(cls.common_words.keys()))
        greeting = cls.time_greeting(language)
        return random.choice((
            greeting,
            greeting.lower(),
            greeting[:-1],
            greeting.lower()[:-1]
        ))

    @classmethod
    def time_greeting(cls, language='es') -> str:
        hour = datetime.datetime.now().hour

        if 7 <= hour < 12:
            return cls.get('greetings', language)[0]
        elif 12 <= hour < 21:
            return cls.get('greetings', language)[1]
        else:
            return cls.get('greetings', language)[2]



from flanautils.asyncs import *
from flanautils.constants import *
from flanautils.data_structures import *
from flanautils.decorators import *
from flanautils.exceptions import *
from flanautils.iterables import *
from flanautils.maths import *
from flanautils.medias import *
from flanautils.models import *
from flanautils.oss import *
from flanautils.strings import *



import asyncio
import datetime
import functools
import html
from asyncio import Task
from typing import Callable, Iterable, Type

import aiohttp
import aiohttp.client_exceptions

from flanautils.exceptions import ResponseError


async def do_every(
    seconds: int | datetime.timedelta,
    func: Callable,
    *args,
    exceptions_to_ignore: Type[BaseException] | Iterable[Type[BaseException]] = (),
    times: int = None,
    while_: bool = True,
    **kwargs
) -> Task:
    """
    Corutine function that runs the function every time provided in seconds or datetime.timedelta.

    You can specify a number of times and a stop condition.

    Exceptions specified in exceptions_to_ignore are ignored if raised.
    """

    if isinstance(seconds, datetime.timedelta):
        seconds = seconds.total_seconds()
    if not isinstance(exceptions_to_ignore, Iterable):
        exceptions_to_ignore = (exceptions_to_ignore,)

    async def do_every_():
        count = 0
        while while_ and (times is None or count < times):
            try:
                await func(*args, **kwargs)
            except (*exceptions_to_ignore,):
                pass
            count += 1
            await asyncio.sleep(seconds)

    return asyncio.create_task(do_every_())


async def do_later(
    seconds: int | datetime.timedelta,
    func: Callable,
    *args,
    exceptions_to_ignore: Type[BaseException] | Iterable[Type[BaseException]] = (),
    **kwargs
) -> Task:
    """
    Corutine function that executes the function after a time provided in seconds or datetime.timedelta.

    Exceptions specified in exceptions_to_ignore are ignored if raised.
    """

    if isinstance(seconds, datetime.timedelta):
        seconds = seconds.total_seconds()
    if not isinstance(exceptions_to_ignore, Iterable):
        exceptions_to_ignore = (exceptions_to_ignore,)

    async def do_later_():
        await asyncio.sleep(seconds)
        try:
            return await func(*args, **kwargs)
        except (*exceptions_to_ignore,):
            pass

    return asyncio.create_task(do_later_())


async def request(http_method, url: str, params: dict = None, headers: dict = None, data: dict = None, session: aiohttp.ClientSession = None, return_response=False, intents=5) -> bytes | str | list | dict | aiohttp.ClientResponse:
    """
    Function that simplifies asynchronous http requests with aiohttp.

    If return_response=True it returns the response object instead of the response data (by default return_response=False).

    Retry the request if it fails up to the number of times specified by intents (by default intents=5).

    Raise exceptions.ResponseError if response.status != 200.
    """

    session_ = session or aiohttp.ClientSession()

    if http_method == 'get':
        http_method = session_.get
    elif http_method == 'post':
        http_method = session_.post
    else:
        raise ValueError('Bad http method.')

    try:
        if params:
            params = {str(k): str(v) for k, v in params.items()}

        for intent in range(intents):
            try:
                async with http_method(url, params=params, headers=headers, data=data) as response:
                    if return_response:
                        return response
                    if response.status != 200:
                        raise ResponseError(f'{response.status} - {response.reason} - {await response.read()}')

                    if response.content_type == 'application/json':
                        return await response.json()
                    elif 'text' in response.content_type:
                        return html.unescape((await response.read()).decode('unicode_escape'))
                    else:
                        return await response.read()
            except aiohttp.client_exceptions.ServerDisconnectedError:
                if intent == 4:
                    raise
                await asyncio.sleep(1)
    finally:
        if not session:
            await session_.close()


get_request = functools.partial(request, 'get')
post_request = functools.partial(request, 'post')



from flanautils.models.bases import *
from flanautils.models.enums import *
from flanautils.models.media import *
from flanautils.models.plotly_charts import *
from flanautils.models.ratio_match import *
from flanautils.models.time_units import *



import asyncio
import platform
import subprocess


async def mp4_to_gif(bytes_: bytes) -> bytes:
    """Convert video in mp4 format given in bytes into video in gif format using FFmpeg."""

    kwargs = {'creationflags': subprocess.CREATE_NO_WINDOW} if platform.system() == 'Windows' else {}
    process = await asyncio.create_subprocess_shell(
        'ffmpeg -i pipe: -f gif -vf "fps=30,split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse" -loop 0 -y pipe:',
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        **kwargs
    )
    stdout, _ = await process.communicate(bytes_)

    return stdout



from dataclasses import dataclass
from typing import Generic, TypeVar

from flanautils.models.bases import FlanaBase

T = TypeVar('T')


@dataclass(frozen=True)
class RatioMatch(FlanaBase, Generic[T]):
    """Dataclass to store the relationship between an element and a punctuation or ratio."""

    element: T
    ratio: float

    def __add__(self, other):
        other_value = other.ratio if isinstance(other, RatioMatch) else other
        return self.ratio + other_value

    def __radd__(self, other):
        return self + other

    def __eq__(self, other):
        if isinstance(other, RatioMatch):
            return (self.element, self.ratio) == (other.element, other.ratio)
        else:
            return self.element == other

    def __lt__(self, other):
        return (other.ratio, str(self.element)) < (self.ratio, str(other.element))
